import json
import uuid
from functools import partial
from http import HTTPStatus
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

import structlog
from anyio import TASK_STATUS_IGNORED, Event, Lock, create_task_group
from anyio.abc import TaskStatus
from fastapi import HTTPException, Response
from fastapi.responses import FileResponse
from starlette.requests import Request
from watchfiles import Change, awatch

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.kernels import Kernels, KernelsConfig
from jupyverse_api.kernels.models import CreateSession, Execution, Kernel, Notebook, Session
from jupyverse_api.main import Lifespan
from jupyverse_api.yjs import Yjs

from .kernel_driver.driver import KernelDriver
from .kernel_driver.kernelspec import find_kernelspec, kernelspec_dirs
from .kernel_driver.paths import jupyter_runtime_dir
from .kernel_server.server import (
    AcceptedWebSocket,
    KernelServer,
    kernels,
)

logger = structlog.get_logger()


class _Kernels(Kernels):
    def __init__(
        self,
        app: App,
        kernels_config: KernelsConfig,
        auth: Auth,
        frontend_config: FrontendConfig,
        yjs: Optional[Yjs],
        lifespan: Lifespan,
    ) -> None:
        super().__init__(app=app, auth=auth)
        self.kernels_config = kernels_config
        self.frontend_config = frontend_config
        self.yjs = yjs
        self.lifespan = lifespan

        self.kernelspecs: dict = {}
        self.kernel_id_to_connection_file: Dict[str, str] = {}
        self.sessions: Dict[str, Session] = {}
        self.kernels = kernels
        self._app = app
        self.stop_event = Event()
        self._stop_lock = Lock()

    async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        async with create_task_group() as tg:
            self.task_group = tg
            if self.kernels_config.allow_external_kernels:
                external_connection_dir = self.kernels_config.external_connection_dir
                if external_connection_dir is None:
                    path = Path(jupyter_runtime_dir()) / "external_kernels"
                else:
                    path = Path(external_connection_dir)
                tg.start_soon(lambda: self.watch_connection_files(path))
            tg.start_soon(self.on_shutdown)
            task_status.started()
            await self.stop_event.wait()

    async def stop(self) -> None:
        async with self._stop_lock:
            if self.stop_event.is_set():
                return

            async with create_task_group() as tg:
                for kernel_id, kernel in self.kernels.items():
                    logger.info("Stopping kernel", kernel_id=kernel_id)
                    tg.start_soon(kernel["server"].stop)
                    if kernel["driver"] is not None:
                        tg.start_soon(kernel["driver"].stop)
            self.stop_event.set()
            self.task_group.cancel_scope.cancel()

    async def on_shutdown(self):
        await self.lifespan.shutdown_request.wait()
        await self.stop()

    async def get_status(
        self,
        user: User,
    ):
        started = self._app.started_time.isoformat().replace("+00:00", "Z")
        last_activity = self._app.last_activity.isoformat().replace("+00:00", "Z")
        connections = sum(
            [kernel["server"].connections for kernel in kernels.values() if "server" in kernel]
        )
        return {
            "started": started,
            "last_activity": last_activity,
            "kernels": len(kernels),
            "connections": connections,
        }

    async def get_kernelspecs(
        self,
        user: User,
    ):
        for search_path in kernelspec_dirs():
            for path in Path(search_path).glob("*/kernel.json"):
                with open(path) as f:
                    spec = json.load(f)
                name = path.parent.name
                resources = {
                    f.stem: f"{self.frontend_config.base_url}kernelspecs/{name}/{f.name}"
                    for f in path.parent.iterdir()
                    if f.is_file() and f.name != "kernel.json"
                }
                self.kernelspecs[name] = {"name": name, "spec": spec, "resources": resources}
        return {"default": self.kernels_config.default_kernel, "kernelspecs": self.kernelspecs}

    async def get_kernelspec(
        self,
        kernel_name,
        file_name,
        user: User,
    ):
        for search_path in kernelspec_dirs():
            file_path = Path(search_path) / kernel_name / file_name
            if file_path.exists():
                return FileResponse(file_path)

        raise HTTPException(
            status_code=404, detail=f"Kernelspec {kernel_name}/{file_name} not found"
        )

    async def get_kernels(
        self,
        user: User,
    ):
        results = []
        for kernel_id, kernel in kernels.items():
            if kernel["server"]:
                connections = kernel["server"].connections
                last_activity = kernel["server"].last_activity["date"]
                execution_state = kernel["server"].last_activity["execution_state"]
            else:
                connections = 0
                last_activity = ""
                execution_state = "idle"
            results.append(
                {
                    "id": kernel_id,
                    "name": kernel["name"],
                    "connections": connections,
                    "last_activity": last_activity,
                    "execution_state": execution_state,
                }
            )
        return results

    async def delete_session(
        self,
        session_id: str,
        user: User,
    ):
        kernel_id = self.sessions[session_id].kernel.id
        kernel_server = kernels[kernel_id]["server"]
        await kernel_server.stop()
        del kernels[kernel_id]
        if kernel_id in self.kernel_id_to_connection_file:
            del self.kernel_id_to_connection_file[kernel_id]
        del self.sessions[session_id]
        return Response(status_code=HTTPStatus.NO_CONTENT.value)

    async def rename_session(
        self,
        request: Request,
        user: User,
    ):
        rename_session = await request.json()
        session_id = rename_session.pop("id")
        for key, value in rename_session.items():
            setattr(self.sessions[session_id], key, value)
        return self.sessions[session_id]

    async def get_sessions(
        self,
        user: User,
    ):
        for session in self.sessions.values():
            kernel_id = session.kernel.id
            if kernel_id in kernels:
                kernel_server = kernels[kernel_id]["server"]
                session.kernel.last_activity = kernel_server.last_activity["date"]
                session.kernel.execution_state = kernel_server.last_activity["execution_state"]
        return list(self.sessions.values())

    async def create_session(
        self,
        request: Request,
        user: User,
    ):
        create_session = CreateSession(**(await request.json()))
        kernel_id = create_session.kernel.id
        kernel_name = create_session.kernel.name
        if kernel_name is not None:
            # launch a new ("internal") kernel
            kernel_cwd = Path(create_session.path).parent
            while True:
                if kernel_cwd.is_dir():
                    break
                kernel_cwd = kernel_cwd.parent
            kernel_server = KernelServer(
                kernelspec_path=Path(find_kernelspec(kernel_name)).as_posix(),
                kernel_cwd=str(kernel_cwd),
            )
            kernel_id = str(uuid.uuid4())
            kernels[kernel_id] = {"name": kernel_name, "server": kernel_server, "driver": None}
            logger.info("Starting kernel", kernel_id=kernel_id, kernel_name=kernel_name)
            await self.task_group.start(kernel_server.start)
        elif kernel_id is not None:
            # external kernel
            kernel_name = kernels[kernel_id]["name"]
            kernel_server = KernelServer(
                connection_file=self.kernel_id_to_connection_file[kernel_id],
                write_connection_file=False,
            )
            kernels[kernel_id]["server"] = kernel_server
            await self.task_group.start(partial(kernel_server.start, launch_kernel=False))
        else:
            return
        session_id = str(uuid.uuid4())
        session = Session(
            id=session_id,
            path=create_session.path,
            name=create_session.name,
            type=create_session.type,
            kernel=Kernel(
                id=kernel_id,
                name=kernel_name,
                connections=kernel_server.connections,
                last_activity=kernel_server.last_activity["date"],
                execution_state=kernel_server.last_activity["execution_state"],
            ),
            notebook=Notebook(
                path=create_session.path,
                name=create_session.name,
            ),
        )
        self.sessions[session_id] = session
        return session

    async def interrupt_kernel(
        self,
        kernel_id,
        user: User,
    ):
        if kernel_id in kernels:
            kernel = kernels[kernel_id]
            kernel["server"].interrupt()
            result = {
                "id": kernel_id,
                "name": kernel["name"],
                "connections": kernel["server"].connections,
                "last_activity": kernel["server"].last_activity["date"],
                "execution_state": kernel["server"].last_activity["execution_state"],
            }
            return result

    async def restart_kernel(
        self,
        kernel_id,
        user: User,
    ):
        if kernel_id in kernels:
            kernel = kernels[kernel_id]
            await self.task_group.start(kernel["server"].restart)
            result = {
                "id": kernel_id,
                "name": kernel["name"],
                "connections": kernel["server"].connections,
                "last_activity": kernel["server"].last_activity["date"],
                "execution_state": kernel["server"].last_activity["execution_state"],
            }
            return result

    async def execute_cell(
        self,
        request: Request,
        kernel_id,
        user: User,
    ):
        if self.yjs is None:
            raise RuntimeError("Cannot execute without a Yjs plugin.")

        r = await request.json()
        execution = Execution(**r)
        if kernel_id in kernels:
            ynotebook = self.yjs.get_document(execution.document_id)
            ycells = [ycell for ycell in ynotebook.ycells if ycell["id"] == execution.cell_id]
            if not ycells:
                return  # FIXME

            ycell = ycells[0]
            del ycell["outputs"][:]

            kernel = kernels[kernel_id]
            if not kernel["driver"]:
                kernel["driver"] = driver = KernelDriver(
                    kernelspec_path=Path(find_kernelspec(kernel["name"])).as_posix(),
                    write_connection_file=False,
                    connection_file=kernel["server"].connection_file_path,
                    yjs=self.yjs,
                )
                await self.task_group.start(driver.start)
            driver = kernel["driver"]

            await driver.execute(ycell, wait_for_executed=False)

    async def get_kernel(
        self,
        kernel_id,
        user: User,
    ):
        if kernel_id in kernels:
            kernel = kernels[kernel_id]
            result = {
                "id": kernel_id,
                "name": kernel["name"],
                "connections": kernel["server"].connections,
                "last_activity": kernel["server"].last_activity["date"],
                "execution_state": kernel["server"].last_activity["execution_state"],
            }
            return result

    async def shutdown_kernel(
        self,
        kernel_id,
        user: User,
    ):
        logger.info("Stopping kernel", kernel_id=kernel_id)
        if kernel_id in kernels:
            await kernels[kernel_id]["server"].stop()
            del kernels[kernel_id]
        for session_id in [k for k, v in self.sessions.items() if v.kernel.id == kernel_id]:
            del self.sessions[session_id]
        return Response(status_code=HTTPStatus.NO_CONTENT.value)

    async def kernel_channels(
        self,
        kernel_id,
        session_id,
        websocket_permissions,
    ):
        if websocket_permissions is None:
            return
        websocket, permissions = websocket_permissions
        subprotocol = (
            "v1.kernel.websocket.jupyter.org"
            if "v1.kernel.websocket.jupyter.org" in websocket["subprotocols"]
            else None
        )
        await websocket.accept(subprotocol=subprotocol)
        accepted_websocket = AcceptedWebSocket(websocket, subprotocol)
        if kernel_id in kernels:
            kernel_server = kernels[kernel_id]["server"]
            if kernel_server is None:
                # this is an external kernel
                # kernel is already launched, just start a kernel server
                kernel_server = KernelServer(
                    connection_file=self.kernel_id_to_connection_file[kernel_id],
                    write_connection_file=False,
                )
                await self.task_group.start(partial(kernel_server.start, launch_kernel=False))
                kernels[kernel_id]["server"] = kernel_server
            await kernel_server.serve(
                accepted_websocket, session_id, permissions, self.lifespan.shutdown_request
            )

    async def watch_connection_files(self, path: Path) -> None:
        # first time scan, treat everything as added files
        initial_changes = {(Change.added, str(p)) for p in path.iterdir()}
        await self.process_connection_files(initial_changes)
        # then, on every change
        async for changes in awatch(path, stop_event=self.stop_event):
            await self.process_connection_files(changes)

    async def process_connection_files(self, changes: Set[Tuple[Change, str]]):
        # get rid of "simultaneously" added/deleted files
        file_changes: Dict[str, List[Change]] = {}
        for c in changes:
            change, path = c
            if path not in file_changes:
                file_changes[path] = []
            file_changes[path].append(change)
        to_delete: List[str] = []
        for p, cs in file_changes.items():
            if Change.added in cs and Change.deleted in cs:
                cs.remove(Change.added)
                cs.remove(Change.deleted)
                if not cs:
                    to_delete.append(p)
        for p in to_delete:
            del file_changes[p]
        # process file changes
        for path, cs in file_changes.items():
            for change in cs:
                if change == Change.deleted:
                    if path in self.kernel_id_to_connection_file.values():
                        kernel_id = list(self.kernel_id_to_connection_file.keys())[
                            list(self.kernel_id_to_connection_file.values()).index(path)
                        ]
                        del self.kernel_id_to_connection_file[kernel_id]
                        del kernels[kernel_id]
                elif change == Change.added:
                    try:
                        data = json.loads(Path(path).read_text())
                    except Exception:
                        continue
                    if "kernel_name" not in data or "key" not in data:
                        continue
                    # looks like a kernel connection file
                    kernel_id = str(uuid.uuid4())
                    self.kernel_id_to_connection_file[kernel_id] = path
                    kernels[kernel_id] = {
                        "name": data["kernel_name"],
                        "server": None,
                        "driver": None,
                    }



from fps import Module

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, AuthConfig

from .config import _AuthFiefConfig
from .routes import auth_factory


class AuthFiefModule(Module):
    def __init__(self, name: str, **kwargs):
        self.auth_fief_config = _AuthFiefConfig(**kwargs)

    async def prepare(self) -> None:
        self.put(self.auth_fief_config, AuthConfig)

        app = await self.get(App)

        auth_fief = auth_factory(app, self.auth_fief_config)
        self.put(auth_fief, Auth)



import os
import signal
import subprocess
import time
from pathlib import Path
from socket import socket

import pytest
import requests


@pytest.fixture
def anyio_backend():
    # at least, SQLAlchemy and pyzmq don't support anything else than asyncio
    # TODO: switch to zmq-anyio
    return "asyncio"


@pytest.fixture()
def cwd():
    return Path(__file__).parents[1]


@pytest.fixture()
def unused_tcp_port() -> int:
    with socket() as sock:
        sock.bind(("127.0.0.1", 0))
        return sock.getsockname()[1]


@pytest.fixture()
def start_jupyverse(auth_mode, clear_users, cwd, unused_tcp_port):
    os.chdir(cwd)
    command_list = [
        "jupyverse",
        "--set",
        f"auth.mode={auth_mode}",
        "--set",
        f"auth.clear_users={str(clear_users).lower()}",
        "--set",
        "kernels.require_yjs=true",
        "--port",
        str(unused_tcp_port),
    ]
    p = subprocess.Popen(command_list)
    url = f"http://127.0.0.1:{unused_tcp_port}"
    while True:
        try:
            requests.get(url)
        except requests.exceptions.ConnectionError:
            time.sleep(0.1)
        else:
            break
    yield url
    os.kill(p.pid, signal.SIGINT)
    p.wait()



import json
from functools import partial
from pathlib import Path

import anyio
import pytest
import requests
from fps_yjs.ydocs import ydocs
from fps_yjs.ywebsocket import WebsocketProvider
from httpx_ws import aconnect_ws
from pycrdt import Array, Doc, Map, Text
from utils import Websocket

prev_theme = {}
test_theme = {"raw": '{// jupyverse test\n"theme": "JupyterLab Dark"}'}


@pytest.mark.parametrize("auth_mode", ("noauth",))
@pytest.mark.parametrize("clear_users", (False,))
def test_settings_persistence_put(start_jupyverse):
    url = start_jupyverse
    # get previous theme
    response = requests.get(url + "/lab/api/settings/@jupyterlab/apputils-extension:themes")
    assert response.status_code == 200
    prev_theme["raw"] = json.loads(response.content)["raw"]
    # put new theme
    response = requests.put(
        url + "/lab/api/settings/@jupyterlab/apputils-extension:themes", data=json.dumps(test_theme)
    )
    assert response.status_code == 204


@pytest.mark.parametrize("auth_mode", ("noauth",))
@pytest.mark.parametrize("clear_users", (False,))
def test_settings_persistence_get(start_jupyverse):
    url = start_jupyverse
    # get new theme
    response = requests.get(
        url + "/lab/api/settings/@jupyterlab/apputils-extension:themes",
    )
    assert response.status_code == 200
    assert json.loads(response.content)["raw"] == test_theme["raw"]
    # put previous theme back
    response = requests.put(
        url + "/lab/api/settings/@jupyterlab/apputils-extension:themes",
        data=json.dumps(prev_theme),
    )
    assert response.status_code == 204


@pytest.mark.anyio
@pytest.mark.parametrize("auth_mode", ("noauth",))
@pytest.mark.parametrize("clear_users", (False,))
async def test_rest_api(start_jupyverse):
    url = start_jupyverse
    name = "notebook0.ipynb"
    path = (Path("tests") / "data" / name).as_posix()
    # create a session to launch a kernel
    response = requests.post(
        f"{url}/api/sessions",
        data=json.dumps(
            {
                "kernel": {"name": "python3"},
                "name": name,
                "path": path,
                "type": "notebook",
            }
        ),
    )
    r = response.json()
    kernel_id = r["kernel"]["id"]
    # get the room ID for the document
    response = requests.put(
        f"{url}/api/collaboration/session/{path}",
        data=json.dumps(
            {
                "format": "json",
                "type": "notebook",
            }
        ),
    )
    file_id = response.json()["fileId"]
    document_id = f"json:notebook:{file_id}"
    ydoc = Doc()
    async with aconnect_ws(
        f"{url}/api/collaboration/room/{document_id}"
    ) as websocket, WebsocketProvider(ydoc, Websocket(websocket, document_id)):
        # connect to the shared notebook document
        # wait for file to be loaded and Y model to be created in server and client
        await anyio.sleep(0.5)
        ydoc["cells"] = ycells = Array()
        # execute notebook
        for cell_idx in range(3):
            response = requests.post(
                f"{url}/api/kernels/{kernel_id}/execute",
                data=json.dumps(
                    {
                        "document_id": document_id,
                        "cell_id": ycells[cell_idx]["id"],
                    }
                ),
            )
        # wait for Y model to be updated
        await anyio.sleep(0.5)
        # retrieve cells
        cells = json.loads(str(ycells))
        assert cells[0]["outputs"] == [
            {
                "data": {"text/plain": "3"},
                "execution_count": 1,
                "metadata": {},
                "output_type": "execute_result",
            }
        ]
        assert cells[1]["outputs"] == [
            {"name": "stdout", "output_type": "stream", "text": ["Hello World!\n"]}
        ]
        assert cells[2]["outputs"] == [
            {
                "data": {"text/plain": "7"},
                "execution_count": 3,
                "metadata": {},
                "output_type": "execute_result",
            }
        ]


@pytest.mark.anyio
@pytest.mark.parametrize("auth_mode", ("noauth",))
@pytest.mark.parametrize("clear_users", (False,))
async def test_ywidgets(start_jupyverse):
    url = start_jupyverse
    name = "notebook1.ipynb"
    path = (Path("tests") / "data" / name).as_posix()
    # create a session to launch a kernel
    response = requests.post(
        f"{url}/api/sessions",
        data=json.dumps(
            {
                "kernel": {"name": "python3"},
                #"kernel": {"name": "akernel"},
                "name": name,
                "path": path,
                "type": "notebook",
            }
        ),
    )
    r = response.json()
    kernel_id = r["kernel"]["id"]
    # get the room ID for the document
    response = requests.put(
        f"{url}/api/collaboration/session/{path}",
        data=json.dumps(
            {
                "format": "json",
                "type": "notebook",
            }
        ),
    )
    file_id = response.json()["fileId"]
    document_id = f"json:notebook:{file_id}"
    ynb = ydocs["notebook"]()
    def callback(aevent, events, event):
        events.append(event)
        aevent.set()
    aevent = anyio.Event()
    events = []
    ynb.ydoc.observe_subdocs(partial(callback, aevent, events))
    async with aconnect_ws(
        f"{url}/api/collaboration/room/{document_id}"
    ) as websocket, WebsocketProvider(ynb.ydoc, Websocket(websocket, document_id)):
        # connect to the shared notebook document
        # wait for file to be loaded and Y model to be created in server and client
        await anyio.sleep(0.5)
        # execute notebook
        for cell_idx in range(2):
            response = requests.post(
                f"{url}/api/kernels/{kernel_id}/execute",
                data=json.dumps(
                    {
                        "document_id": document_id,
                        "cell_id": ynb.ycells[cell_idx]["id"],
                    }
                ),
            )
        while True:
            await aevent.wait()
            aevent = anyio.Event()
            guid = None
            for event in events:
                if event.added:
                    guid = event.added[0]
            if guid is not None:
                break
        async with anyio.create_task_group() as tg:
            tg.start_soon(connect_ywidget, url, guid)
            response = requests.post(
                f"{url}/api/kernels/{kernel_id}/execute",
                data=json.dumps(
                    {
                        "document_id": document_id,
                        "cell_id": ynb.ycells[2]["id"],
                    }
                ),
            )


async def connect_ywidget(url, guid):
    ywidget_doc = Doc()
    async with aconnect_ws(
        f"{url}/api/collaboration/room/ywidget:{guid}"
    ) as websocket, WebsocketProvider(ywidget_doc, Websocket(websocket, guid)):
        await anyio.sleep(0.5)
        attrs = Map()
        model_name = Text()
        ywidget_doc["_attrs"] = attrs
        ywidget_doc["_model_name"] = model_name
        assert str(model_name) == "Switch"
        assert str(attrs) == '{"value":true}'



import json

import pytest
from fps import get_root_module, merge_config
from httpx import AsyncClient

test_theme = {"raw": '{// jupyverse test\n"theme": "JupyterLab Dark"}'}

CONFIG = {
    "jupyverse": {
        "type": "jupyverse",
        "modules": {
            "app": {
                "type": "app",
            },
            "auth": {
                "type": "auth",
                "config": {
                    "test": True,
                },
            },
            "contents": {
                "type": "contents",
            },
            "frontend": {
                "type": "frontend",
            },
            "lab": {
                "type": "lab",
            },
            "jupyterlab": {
                "type": "jupyterlab",
            },
            "kernels": {
                "type": "kernels",
            },
            "yjs": {
                "type": "yjs",
            },
        }
    }
}


@pytest.mark.anyio
@pytest.mark.parametrize("auth_mode", ("noauth",))
async def test_settings(auth_mode, unused_tcp_port):
    config = merge_config(
        CONFIG,
        {
            "jupyverse": {
                "config": {"port": unused_tcp_port},
                "modules": {
                    "auth": {
                        "config": {
                            "mode": auth_mode,
                        }
                    }
                }
            }
        }
    )
    async with get_root_module(config), AsyncClient() as http:
        # get previous theme
        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}/lab/api/settings/@jupyterlab/apputils-extension:themes"
        )
        assert response.status_code == 200
        theme = {"raw": json.loads(response.content)["raw"]}
        # put new theme
        response = await http.put(
            f"http://127.0.0.1:{unused_tcp_port}/lab/api/settings/@jupyterlab/apputils-extension:themes",
            data=json.dumps(test_theme),
        )
        assert response.status_code == 204
        # get new theme
        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}/lab/api/settings/@jupyterlab/apputils-extension:themes"
        )
        assert response.status_code == 200
        assert json.loads(response.content)["raw"] == test_theme["raw"]
        # put previous theme back
        response = await http.put(
            f"http://127.0.0.1:{unused_tcp_port}/lab/api/settings/@jupyterlab/apputils-extension:themes",
            data=json.dumps(theme),
        )
        assert response.status_code == 204



import structlog
from fastapi_users.exceptions import UserAlreadyExists
from fps import Module

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, AuthConfig
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.main import QueryParams

from .config import _AuthConfig
from .routes import auth_factory

log = structlog.get_logger()


class AuthModule(Module):
    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.auth_config = _AuthConfig(**kwargs)

    async def prepare(self) -> None:
        self.put(self.auth_config, AuthConfig)

        app = await self.get(App)
        frontend_config = await self.get(FrontendConfig)

        auth = auth_factory(app, self.auth_config, frontend_config)
        self.put(auth, Auth)

        await auth.db.create_db_and_tables()

        if self.auth_config.test:
            try:
                await auth.create_user(
                    username="admin@jupyter.com",
                    email="admin@jupyter.com",
                    password="jupyverse",
                    permissions={"admin": ["read", "write"]},
                )
            except UserAlreadyExists:
                pass

        try:
            await auth.create_user(
                username=self.auth_config.token,
                email=self.auth_config.global_email,
                password="",
                permissions={},
            )
        except UserAlreadyExists:
            global_user = await auth.get_user_by_email(self.auth_config.global_email)
            await auth._update_user(
                global_user,
                username=self.auth_config.token,
                permissions={},
            )

        if self.auth_config.mode == "token":
            query_params = await self.get(QueryParams)
            query_params.d["token"] = self.auth_config.token



import json
from pathlib import Path

import notebook  # type: ignore
from fastapi.staticfiles import StaticFiles

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.lab import Lab
from jupyverse_api.notebook import Notebook


class _Notebook(Notebook):
    def __init__(self, app: App, auth: Auth, frontend_config: FrontendConfig, lab: Lab) -> None:
        super().__init__(app, auth, lab)
        self.frontend_config = frontend_config
        self.lab = lab
        lab.redirect_after_root = "tree"

        extensions_dir = lab.prefix_dir / "share" / "jupyter" / "labextensions"
        self.federated_extensions, self.disabled_extensions = lab.get_federated_extensions(
            extensions_dir
        )
        self.notebook_dir = Path(notebook.__file__).parent

        self.mount(
            "/static/notebook",
            StaticFiles(directory=self.notebook_dir / "static"),
            name="static",
        )

    async def get_tree(
        self,
        user: User,
    ):
        return get_index(
            self.lab,
            self.notebook_dir,
            self.federated_extensions,
            self.disabled_extensions,
            "Tree",
            "tree",
            self.frontend_config.collaborative,
            self.frontend_config.base_url,
        )

    async def get_notebook(
        self,
        path,
        user: User,
    ):
        return get_index(
            self.lab,
            self.notebook_dir,
            self.federated_extensions,
            self.disabled_extensions,
            path,
            "notebooks",
            self.frontend_config.collaborative,
            self.frontend_config.base_url,
        )

    async def edit_file(
        self,
        path,
        user: User,
    ):
        return get_index(
            self.lab,
            self.notebook_dir,
            self.federated_extensions,
            self.disabled_extensions,
            path,
            "edit",
            self.frontend_config.collaborative,
            self.frontend_config.base_url,
        )

    async def get_console(
        self,
        path,
        user: User,
    ):
        return get_index(
            self.lab,
            self.notebook_dir,
            self.federated_extensions,
            self.disabled_extensions,
            path,
            "consoles",
            self.frontend_config.collaborative,
            self.frontend_config.base_url,
        )

    async def get_terminal(
        self,
        name: str,
        user: User,
    ):
        return get_index(
            self.lab,
            self.notebook_dir,
            self.federated_extensions,
            self.disabled_extensions,
            name,
            "terminals",
            self.frontend_config.collaborative,
            self.frontend_config.base_url,
        )


INDEX_HTML = """\
<!doctype html>
<html>
    <head>
        <meta charset="utf-8"/>
        <meta name="viewport" content="width=device-width,initial-scale=1"/>
        <title>Home</title>
        <link rel="icon" type="image/x-icon" href="/static/favicons/favicon.ico" class="favicon"/>
        <link rel="stylesheet" href="/custom/custom.css"/>
        <script defer="defer" src="FULL_STATIC_URL/main.MAIN_ID.js?v=MAIN_ID"></script>
    </head>
    <body class="jp-ThemedContainer">
        <script id="jupyter-config-data" type="application/json">
            PAGE_CONFIG
        </script>
        <script>
            /* Remove token from URL. */
            (function() {
                var parsedUrl = new URL(window.location.href);
                if (parsedUrl.searchParams.get('token')) {
                    parsedUrl.searchParams.delete('token');
                    window.history.replaceState({}, '', parsedUrl.href);
                }
            }
            )();
        </script>
    </body>
</html>
"""


def get_index(
    lab,
    notebook_dir,
    federated_extensions,
    disabled_extensions,
    doc_name,
    notebook_page,
    collaborative,
    base_url="/",
):
    full_static_url = f"{base_url}static/notebook"
    for path in (notebook_dir / "static").glob("main.*.js"):
        main_id = path.name.split(".")[1]
        break
    page_config = {
        "appName": "Notebook",
        "appNamespace": "notebook",
        "appSettingsDir": (lab.prefix_dir / "share" / "jupyter" / "lab" / "settings").as_posix(),
        "appUrl": "/lab",
        "appVersion": notebook.__version__,
        "baseUrl": base_url,
        "cacheFiles": True,
        "collaborative": collaborative,
        "disabledExtensions": disabled_extensions,
        "extraLabextensionsPath": [],
        "federated_extensions": federated_extensions,
        "frontendUrl": "/notebook/",
        "fullAppUrl": f"{base_url}lab",
        "fullLabextensionsUrl": f"{base_url}lab/extensions",
        "fullLicensesUrl": f"{base_url}lab/api/licenses",
        "fullListingsUrl": f"{base_url}lab/api/listings",
        "fullMathjaxUrl": f"{base_url}static/notebook/components/MathJax/MathJax.js",
        "fullSettingsUrl": f"{base_url}lab/api/settings",
        "fullStaticUrl": f"{base_url}static/notebook",
        "fullThemesUrl": f"{base_url}lab/api/themes",
        "fullTranslationsApiUrl": f"{base_url}lab/api/translations",
        "fullTreeUrl": f"{base_url}lab/tree",
        "fullWorkspacesApiUrl": f"{base_url}lab/api/workspaces",
        "labextensionsPath": [(lab.prefix_dir / "share" / "jupyter" / "labextensions").as_posix()],
        "labextensionsUrl": "/lab/extensions",
        "licensesUrl": "/lab/api/licenses",
        "listingsUrl": "/lab/api/listings",
        "mathjaxConfig": "TeX-AMS-MML_HTMLorMML-full,Safe",
        "notebookLogo": False,
        "notebookPage": notebook_page,
        "schemasDir": (lab.prefix_dir / "share" / "jupyter" / "lab" / "schemas").as_posix(),
        "settingsUrl": "/lab/api/settings",
        "staticDir": (notebook_dir / "static").as_posix(),
        "templatesDir": (notebook_dir / "templates").as_posix(),
        "terminalsAvailable": True,
        "themesDir": (lab.prefix_dir / "share" / "jupyter" / "lab" / "themes").as_posix(),
        "themesUrl": "/lab/api/themes",
        "translationsApiUrl": "/lab/api/translations",
        "treeUrl": "/lab/tree",
        "workspacesApiUrl": "/lab/api/workspaces",
        "wsUrl": "",
    }
    index = (
        INDEX_HTML.replace("PAGE_CONFIG", json.dumps(page_config))
        .replace("MAIN_ID", main_id)
        .replace("DOC_NAME", doc_name)
        .replace("BASE_URL", base_url)
        .replace("FULL_STATIC_URL", full_static_url)
    )
    return index



import pytest
from fastapi import APIRouter
from fps import get_root_module
from httpx import AsyncClient

from jupyverse_api import Router
from jupyverse_api.app import App


@pytest.mark.anyio
@pytest.mark.parametrize(
    "mount_path",
    (
        None,
        "/foo",
    ),
)
async def test_mount_path(mount_path, unused_tcp_port):
    config = {
        "jupyverse": {
            "type": "jupyverse",
            "config": {
                "port": unused_tcp_port,
            },
            "modules": {
                "app": {
                    "type": "app",
                    "config": {
                        "mount_path": mount_path
                    }
                },
            }
        }
    }

    async with AsyncClient() as http, get_root_module(config) as jupyverse_module:
        app = await jupyverse_module.get(App)
        router = APIRouter()

        @router.get("/")
        async def get():
            pass

        Router(app).include_router(router)

        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}")
        expected = 200 if mount_path is None else 404
        assert response.status_code == expected

        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/bar")
        assert response.status_code == 404

        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/foo")
        expected = 404 if mount_path is None else 200



from fps import Module

from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.jupyterlab import JupyterLab, JupyterLabConfig
from jupyverse_api.lab import Lab

from .routes import _JupyterLab


class JupyterLabModule(Module):
    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.jupyterlab_config = JupyterLabConfig(**kwargs)

    async def prepare(self) -> None:
        self.put(self.jupyterlab_config, JupyterLabConfig)

        app = await self.get(App)
        auth = await self.get(Auth)  # type: ignore[type-abstract]
        frontend_config = await self.get(FrontendConfig)
        lab = await self.get(Lab)  # type: ignore[type-abstract]

        jupyterlab = _JupyterLab(app, self.jupyterlab_config, auth, frontend_config, lab)
        self.put(jupyterlab, JupyterLab)



import os
import sys
from pathlib import Path
from time import sleep

import pytest
from anyio import create_task_group
from fps import get_root_module, merge_config
from fps_kernels.kernel_server.server import KernelServer, kernels
from httpx import AsyncClient
from httpx_ws import aconnect_ws

os.environ["PYDEVD_DISABLE_FILE_VALIDATION"] = "1"

CONFIG = {
    "jupyverse": {
        "type": "jupyverse",
        "modules": {
            "app": {
                "type": "app",
            },
            "auth": {
                "type": "auth",
                "config": {
                    "test": True,
                },
            },
            "contents": {
                "type": "contents",
            },
            "frontend": {
                "type": "frontend",
            },
            "lab": {
                "type": "lab",
            },
            "jupyterlab": {
                "type": "jupyterlab",
            },
            "kernels": {
                "type": "kernels",
            },
            "yjs": {
                "type": "yjs",
            },
        }
    }
}


@pytest.mark.anyio
@pytest.mark.parametrize("auth_mode", ("noauth",))
async def test_kernel_messages(auth_mode, capfd, unused_tcp_port):
    kernel_id = "kernel_id_0"
    kernel_name = "python3"
    kernelspec_path = (
        Path(sys.prefix) / "share" / "jupyter" / "kernels" / kernel_name / "kernel.json"
    )
    assert kernelspec_path.exists()
    kernel_server = KernelServer(kernelspec_path=kernelspec_path, capture_kernel_output=False)
    async with create_task_group() as tg:
        await tg.start(kernel_server.start)
        kernels[kernel_id] = {"server": kernel_server}
        msg_id = "0"
        msg = {
            "channel": "shell",
            "parent_header": None,
            "content": None,
            "metadata": None,
            "header": {
                "msg_type": "msg_type_0",
                "msg_id": msg_id,
            },
        }

        config = merge_config(
            CONFIG,
            {
                "jupyverse": {
                    "config": {"port": unused_tcp_port},
                    "modules": {
                        "auth": {
                            "config": {
                                "mode": auth_mode,
                            }
                        }
                    }
                }
            }
        )
        async with get_root_module(config), AsyncClient():
            # block msg_type_0
            msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
            kernel_server.block_messages("msg_type_0")
            async with aconnect_ws(
                f"http://127.0.0.1:{unused_tcp_port}/api/kernels/{kernel_id}/channels?session_id=session_id_0",
            ) as websocket:
                await websocket.send_json(msg)
            sleep(0.5)
            out, err = capfd.readouterr()
            assert not err

            # allow only msg_type_0
            msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
            kernel_server.allow_messages("msg_type_0")
            async with aconnect_ws(
                f"http://127.0.0.1:{unused_tcp_port}/api/kernels/{kernel_id}/channels?session_id=session_id_0",
            ) as websocket:
                await websocket.send_json(msg)
            sleep(0.5)
            out, err = capfd.readouterr()
            assert err.count("[IPKernelApp] WARNING | Unknown message type: 'msg_type_0'") == 1

            # block all messages
            msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
            kernel_server.allow_messages([])
            async with aconnect_ws(
                f"http://127.0.0.1:{unused_tcp_port}/api/kernels/{kernel_id}/channels?session_id=session_id_0",
            ) as websocket:
                await websocket.send_json(msg)
            sleep(0.5)
            out, err = capfd.readouterr()
            assert not err

            # allow all messages
            msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
            kernel_server.allow_messages()
            async with aconnect_ws(
                f"http://127.0.0.1:{unused_tcp_port}/api/kernels/{kernel_id}/channels?session_id=session_id_0",
            ) as websocket:
                await websocket.send_json(msg)
            sleep(0.5)
            out, err = capfd.readouterr()
            assert err.count("[IPKernelApp] WARNING | Unknown message type: 'msg_type_0'") >= 1

            tg.start_soon(kernel_server.stop)



from __future__ import annotations

import structlog
from anyio import create_task_group
from fps import Module

from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.kernels import Kernels, KernelsConfig
from jupyverse_api.main import Lifespan
from jupyverse_api.yjs import Yjs

from .routes import _Kernels

log = structlog.get_logger()


class KernelsModule(Module):
    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.kernels_config = KernelsConfig(**kwargs)

    async def prepare(self) -> None:
        self.put(self.kernels_config, KernelsConfig)

        app = await self.get(App)
        auth = await self.get(Auth)  # type: ignore[type-abstract]
        frontend_config = await self.get(FrontendConfig)
        lifespan = await self.get(Lifespan)
        yjs = (
            await self.get(Yjs)  # type: ignore[type-abstract]
            if self.kernels_config.require_yjs
            else None
        )

        self.kernels = _Kernels(app, self.kernels_config, auth, frontend_config, yjs, lifespan)
        self.put(self.kernels, Kernels)

        async with create_task_group() as tg:
            tg.start_soon(self.kernels.start)
            self.done()

    async def stop(self) -> None:
        await self.kernels.stop()



import os
import time
import uuid
from typing import Any, Dict, Optional, cast

import anyio
from anyio import (
    TASK_STATUS_IGNORED,
    Event,
    create_memory_object_stream,
    create_task_group,
    fail_after,
)
from anyio.abc import TaskGroup, TaskStatus
from anyio.streams.stapled import StapledObjectStream
from pycrdt import Array, Map

from jupyverse_api.yjs import Yjs

from .connect import cfg_t, connect_channel, launch_kernel, read_connection_file
from .connect import write_connection_file as _write_connection_file
from .kernelspec import find_kernelspec
from .message import create_message, receive_message, send_message


def deadline_to_timeout(deadline: float) -> float:
    return max(0, deadline - time.time())


class KernelDriver:
    task_group: TaskGroup

    def __init__(
        self,
        kernel_name: str = "",
        kernelspec_path: str = "",
        kernel_cwd: str = "",
        connection_file: str = "",
        write_connection_file: bool = True,
        capture_kernel_output: bool = True,
        yjs: Optional[Yjs] = None,
    ) -> None:
        self.write_connection_file = write_connection_file
        self.capture_kernel_output = capture_kernel_output
        self.kernelspec_path = kernelspec_path or find_kernelspec(kernel_name)
        self.kernel_cwd = kernel_cwd
        self.yjs = yjs
        if not self.kernelspec_path:
            raise RuntimeError("Could not find a kernel, maybe you forgot to install one?")
        if write_connection_file:
            self.connection_file_path, self.connection_cfg = _write_connection_file(connection_file)
        else:
            self.connection_file_path = connection_file
            self.connection_cfg = read_connection_file(connection_file)
        self.key = cast(str, self.connection_cfg["key"])
        self.session_id = uuid.uuid4().hex
        self.msg_cnt = 0
        self.execute_requests: Dict[str, Dict[str, StapledObjectStream]] = {}
        self.comm_messages: StapledObjectStream = StapledObjectStream(
            *create_memory_object_stream[dict](max_buffer_size=1024)
        )
        self.stopped_event = Event()

    async def restart(self, startup_timeout: float = float("inf")) -> None:
        self.task_group.cancel_scope.cancel()
        await self.stopped_event.wait()
        self.stopped_event = Event()
        async with create_task_group() as tg:
            self.task_group = tg
            msg = create_message("shutdown_request", content={"restart": True})
            await send_message(msg, self.control_channel, self.key, change_date_to_str=True)
            while True:
                msg = cast(
                    Dict[str, Any],
                    await receive_message(self.control_channel, change_str_to_date=True),
                )
                if msg["msg_type"] == "shutdown_reply" and msg["content"]["restart"]:
                    break
            await self._wait_for_ready(startup_timeout)
            self.listen_channels()
            tg.start_soon(self._handle_comms)

    async def start(
        self,
        startup_timeout: float = float("inf"),
        connect: bool = True,
        *,
        task_status: TaskStatus[None] = TASK_STATUS_IGNORED,
    ) -> None:
        async with create_task_group() as tg:
            self.task_group = tg
            self.kernel_process = await launch_kernel(
                self.kernelspec_path,
                self.connection_file_path,
                self.kernel_cwd,
                self.capture_kernel_output,
            )
            if connect:
                await self.connect()
            task_status.started()
        self.stopped_event.set()

    async def connect(self, startup_timeout: float = float("inf")) -> None:
        self.connect_channels()
        await self._wait_for_ready(startup_timeout)
        self.listen_channels()
        self.task_group.start_soon(self._handle_comms)

    def connect_channels(self, connection_cfg: Optional[cfg_t] = None):
        connection_cfg = connection_cfg or self.connection_cfg
        self.shell_channel = connect_channel("shell", connection_cfg)
        self.control_channel = connect_channel("control", connection_cfg)
        self.iopub_channel = connect_channel("iopub", connection_cfg)

    def listen_channels(self):
        self.task_group.start_soon(self.listen_iopub)
        self.task_group.start_soon(self.listen_shell)

    async def stop(self) -> None:
        try:
            self.kernel_process.terminate()
        except ProcessLookupError:
            pass
        await self.kernel_process.wait()
        self.task_group.cancel_scope.cancel()
        if self.write_connection_file:
            path = anyio.Path(self.connection_file_path)
            try:
                await path.unlink()
            except Exception:
                pass

    async def listen_iopub(self):
        while True:
            msg = await receive_message(self.iopub_channel, change_str_to_date=True)
            parent_id = msg["parent_header"].get("msg_id")
            if msg["msg_type"] in ("comm_open", "comm_msg"):
                await self.comm_messages.send(msg)
            elif parent_id in self.execute_requests.keys():
                await self.execute_requests[parent_id]["iopub_msg"].send(msg)

    async def listen_shell(self):
        while True:
            msg = await receive_message(self.shell_channel, change_str_to_date=True)
            msg_id = msg["parent_header"].get("msg_id")
            if msg_id in self.execute_requests.keys():
                await self.execute_requests[msg_id]["shell_msg"].send(msg)

    async def execute(
        self,
        ycell: Map,
        timeout: float = float("inf"),
        msg_id: str = "",
        wait_for_executed: bool = True,
    ) -> None:
        if ycell["cell_type"] != "code":
            return
        ycell["execution_state"] = "busy"
        content = {"code": str(ycell["source"]), "silent": False}
        msg = create_message(
            "execute_request", content, session_id=self.session_id, msg_id=str(self.msg_cnt)
        )
        if msg_id:
            msg["header"]["msg_id"] = msg_id
        else:
            msg_id = msg["header"]["msg_id"]
        self.msg_cnt += 1
        await send_message(msg, self.shell_channel, self.key, change_date_to_str=True)
        self.execute_requests[msg_id] = {
            "iopub_msg": StapledObjectStream(
                *create_memory_object_stream[dict](max_buffer_size=1024)
            ),
            "shell_msg": StapledObjectStream(
                *create_memory_object_stream[dict](max_buffer_size=1024)
            ),
        }
        if wait_for_executed:
            deadline = time.time() + timeout
            while True:
                try:
                    with fail_after(deadline_to_timeout(deadline)):
                        msg = await self.execute_requests[msg_id]["iopub_msg"].receive()
                except TimeoutError:
                    error_message = f"Kernel didn't respond in {timeout} seconds"
                    raise RuntimeError(error_message)
                await self._handle_outputs(ycell["outputs"], msg)
                if (
                    msg["header"]["msg_type"] == "status"
                    and msg["content"]["execution_state"] == "idle"
                ):
                    break
            try:
                with fail_after(deadline_to_timeout(deadline)):
                    msg = await self.execute_requests[msg_id]["shell_msg"].receive()
            except TimeoutError:
                error_message = f"Kernel didn't respond in {timeout} seconds"
                raise RuntimeError(error_message)
            with ycell.doc.transaction():
                ycell["execution_count"] = msg["content"]["execution_count"]
                ycell["execution_state"] = "idle"
            del self.execute_requests[msg_id]
        else:
            self.task_group.start_soon(lambda: self._handle_iopub(msg_id, ycell))

    async def _handle_iopub(self, msg_id: str, ycell: Map) -> None:
        while True:
            msg = await self.execute_requests[msg_id]["iopub_msg"].receive()
            await self._handle_outputs(ycell["outputs"], msg)
            if (
                msg["header"]["msg_type"] == "status"
                and msg["content"]["execution_state"] == "idle"
            ):
                msg = await self.execute_requests[msg_id]["shell_msg"].receive()
                with ycell.doc.transaction():
                    ycell["execution_count"] = msg["content"]["execution_count"]
                    ycell["execution_state"] = "idle"
                break

    async def _handle_comms(self) -> None:
        if self.yjs is None or self.yjs.widgets is None:  # type: ignore
            return

        while True:
            msg = await self.comm_messages.receive()
            msg_type = msg["header"]["msg_type"]
            if msg_type == "comm_open":
                comm_id = msg["content"]["comm_id"]
                comm = Comm(comm_id, self.shell_channel, self.session_id, self.key, self.task_group)
                self.yjs.widgets.comm_open(msg, comm)  # type: ignore
            elif msg_type == "comm_msg":
                self.yjs.widgets.comm_msg(msg)  # type: ignore

    async def _wait_for_ready(self, timeout):
        deadline = time.time() + timeout
        new_timeout = timeout
        while True:
            msg = create_message(
                "kernel_info_request", session_id=self.session_id, msg_id=str(self.msg_cnt)
            )
            self.msg_cnt += 1
            await send_message(msg, self.shell_channel, self.key, change_date_to_str=True)
            msg = await receive_message(
                self.shell_channel, timeout=new_timeout, change_str_to_date=True
            )
            if msg is None:
                error_message = f"Kernel didn't respond in {timeout} seconds"
                raise RuntimeError(error_message)
            if msg["msg_type"] == "kernel_info_reply":
                msg = await receive_message(
                    self.iopub_channel, timeout=0.2, change_str_to_date=True
                )
                if msg is not None:
                    break
            new_timeout = deadline_to_timeout(deadline)

    async def _handle_outputs(self, outputs: Array, msg: Dict[str, Any]):
        msg_type = msg["header"]["msg_type"]
        content = msg["content"]
        if msg_type == "stream":
            with outputs.doc.transaction():
                # TODO: uncomment when changes are made in jupyter-ydoc
                text = content["text"]
                if text.endswith((os.linesep, "\n")):
                    text = text[:-1]
                if (not outputs) or (outputs[-1]["name"] != content["name"]):  # type: ignore
                    outputs.append(
                        Map(
                            {
                                "name": content["name"],
                                "output_type": msg_type,
                                "text": Array([content["text"]]),
                            }
                        )
                    )
                else:
                    outputs[-1]["text"].append(content["text"])  # type: ignore
        elif msg_type in ("display_data", "execute_result"):
            if "application/vnd.jupyter.ywidget-view+json" in content["data"]:
                # this is a collaborative widget
                model_id = content["data"]["application/vnd.jupyter.ywidget-view+json"]["model_id"]
                if self.yjs is not None and self.yjs.widgets is not None:  # type: ignore
                    if model_id in self.yjs.widgets.widgets:  # type: ignore
                        doc = self.yjs.widgets.widgets[model_id]["model"].ydoc  # type: ignore
                        path = f"ywidget:{doc.guid}"
                        await self.yjs.room_manager.websocket_server.get_room(path, ydoc=doc)  # type: ignore
                        outputs.append(doc)
            else:
                output = {
                    "data": content["data"],
                    "metadata": {},
                    "output_type": msg_type,
                }
                if msg_type == "execute_result":
                    output["execution_count"] = content["execution_count"]
                outputs.append(output)
        elif msg_type == "error":
            outputs.append(
                {
                    "ename": content["ename"],
                    "evalue": content["evalue"],
                    "output_type": "error",
                    "traceback": content["traceback"],
                }
            )


class Comm:
    def __init__(
        self, comm_id: str, shell_channel, session_id: str, key: str, task_group: TaskGroup
    ):
        self.comm_id = comm_id
        self.shell_channel = shell_channel
        self.session_id = session_id
        self.key = key
        self.task_group = task_group
        self.msg_cnt = 0

    def send(self, buffers):
        msg = create_message(
            "comm_msg",
            content={"comm_id": self.comm_id},
            session_id=self.session_id,
            msg_id=self.msg_cnt,
            buffers=buffers,
        )
        self.msg_cnt += 1
        self.task_group.start_soon(
            lambda: send_message(msg, self.shell_channel, self.key, change_date_to_str=True)
        )
